/***************************************************************************************************
* Program:    Super Nintendo Entertainment System(tm) Audio Processing Unit Emulator               *
* Platform:   Intel 80386 & MMX                                                                    *
* Programmer: Anti Resonance                                                                       *
*                                                                                                  *
* "SNES" and "Super Nintendo Entertainment System" are trademarks of Nintendo Co., Limited and its *
* subsidiary companies.                                                                            *
*                                                                                                  *
* This program is free software; you can redistribute it and/or modify it under the terms of the   *
* GNU General Public License as published by the Free Software Foundation; either version 2 of     *
* the License, or (at your option) any later version.                                              *
*                                                                                                  *
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;        *
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.        *
* See the GNU General Public License for more details.                                             *
*                                                                                                  *
* You should have received a copy of the GNU General Public License along with this program;       *
* if not, write to the Free Software Foundation, Inc.                                              *
* 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                        *
*                                                                                                  *
* ------------------------------------------------------------------------------------------------ *
* Revision History:                                                                                *
*                                                                                                  *
* 1.5  04.11.2003 SNESAPU 2.0                                                                      *
*      + Updated EmuAPU to use SetEmuDSP when DSPINTEG is enabled                                  *
*      + Move fade out code into DSP.Asm                                                           *
*                                                                                                  *
* 1.41 12.07.2003 SNESAPU 1.0a                                                                     *
*      - Fixed the call to SetDSPVol in SetFade that wasn't restoring the stack                    *
*                                                                                                  *
* 1.4  20.06.2003                                                                                  *
*      + Added LoadSPCFile                                                                         *
*      + Added ability to pass -1 to ResetAPU to keep current amp level                            *
*                                                                                                  *
* 1.3  10.02.2003                                                                                  *
*      + Rewrote in assembly                                                                       *
*      + Removed GetAPUData and memory allocation (they were specific to SNESAPU.DLL)              *
*      - Sometimes EmuAPU would generate an extra sample                                           *
*                                                                                                  *
* 1.22 20.06.2002                                                                                  *
*      - Call SetFade after seeking                                                                *
*                                                                                                  *
* 1.21 12.04.2001                                                                                  *
*      + Updated SeekAPU to incorporate new features in SPC700 and DSP emus                        *
*                                                                                                  *
* 1.2  30.12.2000 SNESAmp v2.0                                                                     *
*      + Added SetAPULength, SeekAPU, SaveAPU, RestoreAPU, HaltAPU                                 *
*      + General optimizations                                                                     *
*      - Simplified interface a bit                                                                *
*                                                                                                  *
* 1.1  04.04.2000 SNESAmp v1.0                                                                     *
*                                                                                                  *
* 1.0  17.03.2000 SNESAmp v0.9                                                                     *
*                                                        Copyright (C)2003-04 Alpha-II Productions *
***************************************************************************************************/

#ifndef	__INC_APU
#define	__INC_APU

//**************************************************************************************************
// Defines

//The build options correspond to the equates at the top of APU.Inc

//SPC700 build options bits 7-0 ----------------
#define	SA_HALFC	0x02						//Half-carry enabled
#define	SA_CNTBK	0x04						//Counter Break
#define	SA_SPEED	0x08						//Speed hack
#define	SA_IPLW		0x10						//IPL ROM region writeable
#define	SA_DSPBK	0x20						//Break SPC700/Update DSP if 0F3h is read from

//DSP build options bits 15-8 ------------------
#define	SA_VMETERM	0x100						//Volume metering on main output (for APR)
#define	SA_VMETERC	0x200						//Volume metering on voices (for visualization)
#define	SA_SNESINT	0x400						//Use pregenerated gaussian curve from SNES
#define	SA_STEREO	0x800						//Stereo controls (seperation and EFBCT)

//APU build options bits 23-16 -----------------
#define	SA_DEBUG	0x10000						//Debugging ability
#define	SA_DSPINTEG	0x20000						//DSP emulation is integrated with the SPC700

////////////////////////////////////////////////////////////////////////////////////////////////////
// Private Declarations

#ifndef	SNESAPU_DLL

#ifdef	__cplusplus
extern "C"	u32	apuOpt;							//Build options
#else
extern u32	apuOpt;
#endif


////////////////////////////////////////////////////////////////////////////////////////////////////
// External Functions

#ifdef	__cplusplus
extern "C" {
#endif

//**************************************************************************************************
// Initialize Audio Processing Unit
//
// Initializes internal tables, mixer settings, memory, and registers.
// This only needs to be called once, and must be proceeded by ResetAPU.
//
// In:
//    pAPURAM -> 64kB of memory aligned on a 64kB boundary

void InitAPU(void *pAPURAM);


//**************************************************************************************************
// Reset Audio Processor
//
// Clears all memory, sets registers to default values, and sets the amplification level.
//
// In:
//   amp = Amplification (-1 = keep current amp level, see SetDSPAmp for more information)

void ResetAPU(u32 amp);


//**************************************************************************************************
// Fix Audio Processor After Load
//
// Prepares the sound processor for emulation after an .SPC/.ZST is loaded.
//
// In:
//    SPC700 internal registers

void FixAPU(u16 pc, u8 a, u8 y, u8 x, u8 psw, u8 sp);


//**************************************************************************************************
// Load SPC File
//
// Restores the APU state from an SPC file.  This eliminates the need to call ResetAPU, copy memory,
// and call FixAPU.
//
// In:
//    pFile -> 66048 byte SPC file

void LoadSPCFile(void *pFile);


//**************************************************************************************************
// Save Audio Processor State
//
// Creates a saved state of the audio processor
//
// Note:  Pointers within the structure can be NULL
//
// In:
//    pSPC -> SPCState structure
//    pDSP -> DSPState structure

void SaveAPU(SPCState *pSPC, DSPState *pDSP);


//**************************************************************************************************
// Restore Audio Processor State
//
// Restores the audio processor from a saved state
//
// Notes:  Pointers within the structure can be NULL
//         Passing a NULL value in pEchoBuf will cause the echo to be reset
//         You do not need to call FixAPU after RestoreAPU
//
// In:
//    pSPC -> SPCState structure
//    pDSP -> DSPState structure

void RestoreAPU(SPCState *pSPC, DSPState *pDSP);


//**************************************************************************************************
// Set Audio Processor Options
//
// Configures the sound processor emulator.  Range checking is performed on all parameters.
//
// Notes:  -1 can be passed for any parameter you want to remain unchanged
//         see SetDSPOpt() in DSP.h for a more detailed explantion of the options
//
// In:
//    mix   = Mixing routine (default MIX_INT)
//    chn   = Number of channels (1 or 2, default 2)
//    bits  = Sample size (8, 16, 24, 32, or -32 [IEEE 754], default 16)
//    rate  = Sample rate (8000-192000, default 32000)
//    inter = Interpolation type (default INT_GAUSS)
//    opts  = See 'DSP options' in the Defines section of DSP.h

void SetAPUOpt(u32 mix, u32 chn, u32 bits, u32 rate, u32 inter, u32 opts);


//**************************************************************************************************
// Set Audio Processor Sample Clock
//
// Calculates the ratio of emulated clock cycles to sample output.  Used to speed up or slow down a
// song without affecting the pitch.
//
// In:
//    speed = Multiplier [16.16] (1/2x to 16x)

void SetAPUSmpClk(u32 speed);


//**************************************************************************************************
// Set Audio Processor Song Length
//
// Sets the length of the song and fade
//
// Notes:  If a song is not playing, you must call ResetAPU or set T64Cnt to 0 before calling this.
//         To set a song with no length, pass -1 and 0 for the song and fade.
//
// In:
//    song = Length of song (in 1/64000ths second)
//    fade = Length of fade (in 1/64000ths second)
//
// Out:
//    Total length

u32 SetAPULength(u32 song, u32 fade);


//**************************************************************************************************
// Emulate Audio Processing Unit
//
// Emulates the APU for a specified amount of time.  DSP output is placed in a buffer to be handled
// by the main program.
//
// In:
//    pBuf-> Buffer to store output samples
//    len  = Length of time to emulate (must be > 0)
//    type = Type of parameter passed in len
//           0 - len is the number of APU clock cycles to emulate (APU_CLK = 1 second)
//           1 - len is the number of samples to generate
// Out:
//    -> End of buffer

void* EmuAPU(void *pBuf, u32 len, u8 type);


//**************************************************************************************************
// Seek to Position
//
// Seeks forward in the song from the current position
//
// In:
//    time = 1/64000ths of a second to seek forward (must be >= 0)
//    fast = Use faster seeking method (may break some songs)

void SeekAPU(u32 time, b8 fast);


//**************************************************************************************************
// Shutdown Audio Processing Unit
//
// Currently does nothing

void ShutAPU();

#ifdef	__cplusplus
}
#endif

#endif	//SNESAPU_DLL
#endif	//__INC_APU
