;
;Program:    SNES Digital Signal Processor Emulator
;Platform:   80386 / MMX
;Programmer: Anti Resonance
;
;"SNES" and "Super Nintendo Entertainment System" are trademarks of Nintendo Co., Limited and its
;subsidiary companies.
;
;This program is free software; you can redistribute it and/or modify it under the terms of the
;GNU General Public License as published by the Free Software Foundation; either version 2 of
;the License, or (at your option) any later version.
;
;This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
;without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
;See the GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License along with this program;
;if not, write to the Free Software Foundation, Inc.
;59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;
;Revision History:
;
; 5.0  26.01.2004 SNESAPU v2.0
;      + Added 24- and 32-bit sample support to MMX
;      + Hardware anomalies emulates the DSP at 32kHz then upsamples the output
;      + Replaced fake low-pass filter with a 32-tap FIR filter
;      + Writing to the DSP registers from the SPC700 automatically calls EmuDSP.  This should help
;        keep the DSP and SPC700 in better syncronization.
;      + Removed self modifying code
;      + Moved fade out from APU.Asm
;      - Output is inverted only when hardware anomalies are enabled
;      - Obtained a more accurate copy of the gaussian look-up table used in the SNES
;      - Gaussian interpolation correctly clears the LSB in MMX mixing mode
;      - When using 4-point interpolation, source playback begins with the fourth sample
;      - Removed 17-bit clamp from filter 1
;      - Envelopes are not based on a 33kHz clock (duh...)
;      - Envelopes now use a sample counter instead of a fractional clock
;      - FIR filter coefficients were getting applied in reverse order
;      - Filtered samples get shifted right before accumulation, and LSB is cleared after accum.
;      - Fixed PckWav (it was seriously broken)
;      - Added support for BRR_??? options to PckWav
;
; 4.26 29.07.2003 SNESAPU v1.01
;      - Voice's inactive flag wasn't getting set after a key off
;
; 4.25 30.06.2003
;      - Made some minor changes to RestoreDSP
;      - Fixed voices' pitch not getting changed in SetDSPOpt
;      - Fixed a bug in the floating-point DSP volume register handler when STEREO = 0
;
; 4.2  18.06.2003 SNESAPU v0.99c
;      - Fixed interger overflow in MMX FIR filter
;
; 4.11 14.06.2003 SNESAPU v0.99b
;      + Added support for old amplification values (<= 256)
;
; 4.1  19.05.2003 SNESAPU v0.99a
;      - Fixed a bug in the i386 FIR filter
;
; 4.0  28.04.2003 SNESAPU v0.99
;      + InitDSP will select 386 or MMX depending on CPU
;      + Added 16 and 24-bit sample output to EmuDSPF
;      + Changed SetDSPAmp to decibels
;      - Fixed several bugs in the envelope code so envelopes now work the same as the SNES (fixes a
;        lot of songs)
;      - Stopped caching loop points (fixes Castlevania IV, which likes to change sources after a
;        key on)
;      - Added clamping to the pitch modulation code (fixes CT)
;      - Fixed a bug in KON/KOF emulation
;      - Changed 16-bit mixing process to emulate the SNES
;      - Improved low-pass filtering and simulated RF interference
;      - Removed floating-point support from SetDSPAmp
;      - Fixed another volume bug in SetDSPOpt when switching between integer and float routines
;
; 3.0  10.02.2003 SNESAPU v0.98
;      + Converted to NASM
;      + Added SetDSPDbg, SaveDSP, RestoreDSP, SetDSPReg
;      + Added a loop to EmuDSP so the number of samples passed in can be any value
;      + Added floating-point support to SetDSPAmp
;      + Added micro ramping to the channel volumes in the floating-point routine
;      + Added 32-bit IEEE 754 output
;      - Added a 17-bit clamp to the UnpckWav (fixes Hiouden; CT and FF6 are iffy now)
;      - Added linear interpolation to FIR filter (fixes some songs that turned to noise)
;      - SetDSPOpt wasn't clearing EAX, which was causing dspDecmp to sometimes contain a bad value
;      - ENVX always gets updated, regardless of ADSR/GAIN flag (fixes Clue)
;
; 2.1  xx.06.2001 SNESAPU v0.96
;      - freqTab was built from the wrong values (Hz instead of samples)
;      - rateTab now uses 32-bits of precision instead of 16
;      - Values in rateTab >= 1.0 are now set to 0.99 to reduce the ill effects of fast envelopes
;        when the sample rate is < 32kHz
;      - Reset all internal pitch values when PMON is written to (fixes Asterix)
;      - Clear bits in KOF if voice is inactive (improves EOS detection)
;      - Switched to a static lookup table generated by the SNES for Gaussian interpolation
;      - Damn near cracked the sample decompression
;      - Optimized code a bit in EmuDSPN and inadvertently fixed a bug that was causing some songs
;        to break after seeking
;      - EmuDSPN wasn't checking the loop counter after an envelope update was performed (seeking
;        no longer appears to hang)
;
; 2.0  15.05.2001 SNESAPU v0.95
;      + Added 4-point Gaussian interpolation
;      + EmuDSPN now decompresses samples, so seeking doesn't lose voices
;      + Added an option to FixSeek to disable resetting voices
;      + Reverse stereo swaps the left/right volumes instead of physically swapping samples
;      + More minor optimizations
;      - Changed the sample decompression routine.  Less songs should have problems.
;      - Echo delay was incorrect for most sample rates
;      - Did something to fix the noise generator not working in some songs
;
; 1.9  02.03.2001 SNESAPU v0.92
;      - Fixed the low-pass filter (I'm amazed it was sort of working at all)
;      - Added a hack to UnpckWav to subtract 3 from the range if > 12
;      - Forgot that the DSP needs to be limited to 64K as well (Secret of Evermore doesn't crash)
;      - Fixed a problem in the MMX dynamic code allocation
;      - FixSeek wasn't resetting the DSP X registers (seeking works in Der Langrisser)
;
; 1.8  26.01.2001 SNESAPU v0.90
;      + Added voice monitor and VMax2dB for visualization
;      + Added SetDSPPitch so people with SB/SBPro's and compatibles can hear the correct tone
;      + Added SetDSPStereo for headphone listeners
;      + Added SetDSPEFBCT to make the echo sound more natural
;      + Added UnpackWave so waveforms could be decompressed by external programs
;      + Added 32-bit floating point mixer for people with 24/32-bit sound cards
;      + Added a reverse stereo option
;      + Added GetProcInfo to detect 3DNow! and SSE capabilites
;      + Added EmuDSP which automatically calls the mixing routine set in SetDSPOpt
;      + Made a lot of minor optimizations to DSPIn and EmuDSP
;      + Made separate register handlers in DSPIn for mono and floating-point
;      - Changed SetDSPOpt to use individual parameters instead of a record
;      - SetDSPOpt performs range checking
;      - Fixed configuration implementation so options don't need to be reset for every song and
;        they can be changed between calls to EmuDSP
;      - Fixed mono related code (sounds better now)
;      - Echo filter was never getting turned on
;      - Implemented a better low-pass filter
;
; 1.4  22.05.2000 SNESAmp v1.2
;      + Added a low-pass filter
;      + Added output monitor
;      + Optimized the waveform code in EmuDSPN
;      + Added FixSeek
;      + SetDSPVol now uses a 16-bit value
;      - FixDSP was marking voices as keyed even if the pitch was 0
;      - Correctly implemented echo volume (should be perfect now)
;
; 1.1  05.04.2000 SNESAmp v1.0
;      - Voices with a pitch of 0 can't be keyed on
;      - ENVX becomes 0 when voice is keyed off
;      - ENVX is set to 0 when envelope is in GAIN mode
;      - Account for sample blocks with a range greater than 16-bits (ADPCM still not correct)
;
; 1.0  17.03.2000 SNESAmp v0.9
;                                                        Copyright (C)1999-2004 Alpha-II Productions
;

%define DSP_INC

;
;Equates

;CPU types (for GetProcType)
CPU_MMX		EQU	1
CPU_3DNOW	EQU	2
CPU_SIMD		EQU	4

;Mixing routines (see source for additional notes)
MIX_NONE		EQU	0									;No mixing
MIX_INT		EQU	1									;Use integer instructions
MIX_FLOAT	EQU	3									;Use floating-point instructions

;Interpolation routines
INT_NONE		EQU	0									;None
INT_LINEAR	EQU	1									;Linear
INT_CUBIC	EQU	2									;Cubic
INT_GAUSS	EQU	3									;Gaussian

;DSP options
DSP_ANALOG	EQU	1									;Simulate analog anomalies
DSP_OLDSMP	EQU	2									;Traditional ADPCM routine
DSP_SURND	EQU	4									;"Surround" sound
DSP_REVERSE	EQU	8									;Reverse stereo samples
DSP_NOECHO	EQU	10h								;Disable echo (must not be bit 0 or 5, see disEcho)

;PackWave options
BRR_LINEAR	EQU	1									;Use linear compression for all blocks
BRR_LOOP		EQU	2									;Set loop flag

;Mixing flags
MFLG_MUTE	EQU	01h								;Mute voice
MFLG_KOFF	EQU	04h								;Voice is in the process of keying off
MFLG_OFF		EQU	08h								;Voice is currently inactive
MFLG_END		EQU	10h								;End block was just played


;
;Structures

;DSP registers ------------------------------
STRUC DSPVoice
	volL	resb	1										;Volume Left
	volR	resb	1										;Volume Right
	pitch	resw	1										;Pitch (Rate/32000) [3.11]
	srcn	resb	1										;Wave form being played back
	adsr	resw	1										;Envelope rates: attack, decay, and sustain
	gain	resb	1										;Envelope gain rate (if not using ADSR)
	envx	resb	1										;Current envelope heigth [.7]
	outx	resb	1										;Current sample being output [-.7]
			resb	2
ENDSTRUC

STRUC DSPRAM
	voice	resb	12										;Settings for voice 0
	mvolL	resb	1										;Main Volume Left [-.7]
	efb	resb	1										;Echo Feedback [-.7]
			resb	1
	fc		resb	1										;Filter Coefficient 0
	v2		resb	12
	mvolR	resb	1										;Main Volume Right [-.7]
			resb	3
	v3		resb	12
	evolL	resb	1										;Echo Volume Left [-.7]
	pmon	resb	1										;Pitch Modulation on/off for each voice
			resb	2
	v4		resb	12
	evolR	resb	1										;Echo Volume Right [-.7]
	non	resb	1										;Noise output on/off for each voice
			resb	2
	v5		resb	12
	kon	resb	1										;Key On for each voice
	eon	resb	1										;Echo on/off for each voice
			resb	2
	v6		resb	12
	kof	resb	1										;Key Off for each voice (instantiates release mode)
	dir	resb	1										;Page containing source directory (wave table offsets)
			resb	2
	v7		resb	12
	flg	resb	1										;DSP flags and noise frequency
	esa	resb	1										;Starting page used to store echo waveform
			resb	2
	v8		resb	12
	endx	resb	1										;Waveform has ended
	edl	resb	1										;Echo Delay in ms >> 4
			resb	2
ENDSTRUC

;Internal mixing settings -------------------
STRUC VoiceMix
	;Voice ------------08
	dspv	resd	1										;-> voice registers in DSP
			resd	1
	;Waveform ---------06
	bCur	resd	1										;-> current block
	bHdr	resb	1										;Block Header for current block
	mFlg	resb	1										;Mixing flags
	;Envelope ---------22
	eMode	resb	1										;[3-0] Current envelope mode
															;[6-4] ADSR mode to switch into from Gain
															;[7]   Envelope is not changing
	eRIdx	resb	1										;Index in rateTab
	eRate	resd	1										;Rate of envelope adjustment [16.16]
	eCnt	resd	1										;Sample counter [16.16]
	eVal	resd	1										;Current envelope value
	eAdj	resd	1										;Amount to adjust envelope height
	eDest	resd	1										;Envelope Destination [16.16]
	;Visualization ----08
	vMax
	vMaxL	resd	1										;Maximum absolute sample output
	vMaxR	resd	1
	;Samples ----------52
	sP1	resd	1										;Last sample decompressed (prev1)
	sP2	resd	1										;Second to last sample (prev2)
	sIdx	resb	4										;-> current sample in sBuf
	sBufP	resw	4										;Last 4 samples from previous block
	sBuf	resw	16										;32 bytes for decompressed samples
	;Mixing -----------32
	mTgtL	resd	1										;Target volume
	mTgtR	resd	1										; "  "
	mChnL	resd	1										;Channel volume [-24.7]
	mChnR	resd	1										; "  "
	mRate	resd	1										;Pitch Rate after modulation [16.16]
	mDec	resw	2										;Pitch Decimal [.16] (used as delta for interpolation)
	mOrgP	resd	1										;Original pitch rate converted from the DSP [16.16]
	mOut	resd	1										;Last sample output before chn vol (used for pitch mod)
ENDSTRUC

;Saved state (see SaveDSP) ------------------
STRUC DSPState
	pDSP		resd	1									;[0.0] -> DSP registers (128 bytes)
	pMix		resd	1									;[0.4] -> Internal mixing settings (1k)
	pEchoBuf	resd	1									;[0.8] -> echo buffer (bytes = sample rate * 1.92)
	_rd1		resd	1									;[0.C] reserved

	vMMaxLS	resd	1									;[1.0] Maximum output so far
	vMMaxRS	resd	1
	mAmp		resd	1									;[1.8] Amplification
	vActive	resb	1									;[1.C] Flags for each active voice
	_rd2		resb	3									;[1.D] reserved
ENDSTRUC


;
;Public Variables

PUBLIC	dsp											;1 page of DSP RAM
PUBLIC	mix											;Mixing structures for each voice
PUBLIC	vMMaxL										;Maximum absolute sample output
PUBLIC	vMMaxR


;
;Exported Functions

;
;Get Processor Information
;
;Detects if processor has MMX, 3DNow!, or SSE capabilities
;
;In:
;   nothing
;
;Out:
;   EAX = see CPU_xxx equates
;
;Destroys:
;   nothing

PUBLIC GetProcInfo, NULL


;
;Initialize DSP
;
;Creates the lookup tables for interpolation, and sets the default mixing settings:
;
;   mixType = MIX_INT
;   numChn  = 2
;   bits    = 16
;   rate    = 32000
;   inter   = INT_GAUSS
;   opts    = 0
;
;In:
;   pAPUBase -> 64KB of APU RAM aligned on a 64K boundary
;
;Destroys:
;   ST0-7

PUBLIC InitDSP, pAPUBase:Ptr


;
;Reset DSP
;
;Resets the DSP registers, erases internal variables, and resets the volume
;
;Destroys:
;   EAX

PUBLIC ResetDSP, NULL


;
;Set DSP Options
;
;Recalculates tables, changes the output sample rate, and sets up the mixing routine
;
;Notes:
;   Range checking is performed on all parameters.  If a parameter does not match the required
;    range of values, the default value will be assumed.
;
;   -1 can be used for any paramater that should remain unchanged.
;
;   The DSP_LOW option only works when:
;    mix  =  MIX_INT
;    chn  =  2
;    bits >= 16
;    rate >= 32000
;
; In:
;   mixType = Mixing routine (default MIX_INT)
;   numChn  = Number of channels (1 or 2, default 2)
;   bits    = Sample size (mono 8 or 16; stereo 8, 16, 24, 32, or -32 [IEEE 754], default 16)
;   rate    = Sample rate (8000-192000, default 32000)
;   inter   = Interpolation type (default INT_GAUSS)
;   opts    = See "DSP options" in the Equates section
;
;Destroys:
;   EAX

PUBLIC SetDSPOpt, mixType:dword, numChn:dword, bits:dword, rate:dword, inter:dword, opts:dword


;
;Debug DSP
;
;Installs a vector that gets called each time a value is written to the DSP data register.
;The function is called with the C calling convention.
;
;Upon entrance to the function:
;   [ESP+4] = Current opcode (low word = PC)
;   [ESP+8] = YA
;
;Note:
;   DEBUG must be set to 1 in APU.Inc
;
;In:
;   pTrace-> debug function (a null pointer turns off the debug call)
;
;Out:
;   Previously installed vector

PUBLIC SetDSPDbg, pTrace:ptr


;
;Fix DSP After Loading SPC file
;
;Initializes the internal mixer variables
;
;Destroys:
;   EAX

PUBLIC FixDSP, NULL


;
;Fix DSP After Seeking
;
;Puts all DSP voices in a key off state and erases echo region.  Call after using EmuDSPN.
;
;In:
;   reset = True, Reset all voices
;           False, Only erase memory
;
;Destroys:
;   EAX

PUBLIC FixSeek, reset:byte


;
;Save DSP State
;
;Notes:
;   Pointers in the DSPState structure can be NULL
;
;   Pointers in the Voice structure are not returned as literal values, but as an index instead:
;    Voice.dspv  = index into APU RAM (0-65535)
;    Voice.bCur  = index into APU RAM (0-65535)
;    Voice.sIdx  = index into sBuf (0-15)
;
;In:
;   pState -> Saved state structure
;
;Destroys:
;   EAX

PUBLIC SaveDSP, pState:ptr


;
;Restore DSP State
;
;Notes:
;   Setting a pointer in DSPState to NULL will keep its corresponding internal data intact, except
;    for the echo.  If pEchoBuf is NULL, the internal echo buffer will be cleared.
;
;   Pointers in the Voice structure must contain index values, not literal pointers (see SaveDSP)
;
;   Because the envelope (e???) and mixing (m???) variables contain values affected by the current
;    sample rate and size, these values may need to be reset.  By setting eRate or mRate to -1,
;    RestoreDSP will recaluclate their respective values.
;
;   When setting eRate to -1, eMode must contain the envelope mode when the state was saved, and
;    eVal must contain the envelope height as a 7-bit quantity (as in ENVX), not a fixed point
;    quantity as it was saved out.
;
;   mFlg is not affected by setting mRate to -1.
;
;In:
;   pState -> Saved state structure
;
;Destroys:
;   EAX

PUBLIC RestoreDSP, pState:ptr


;
;DSP Pitch Adjustment
;
;Adjusts the pitch of the DSP
;
;In:
;   base = Base sample rate (32000 - Normal pitch, 32458 - Old SB cards, 32768 - Old ZSNES)
;
;Destroys:
;   EAX

PUBLIC SetDSPPitch, base:dword


;
;Set DSP Amplification
;
;This value is applied to the output with the main volumes
;
;In:
;   amp = Amplification level [-15.16] (1.0 = SNES, negative values act as 0)
;
;Destroys:
;   EAX

PUBLIC SetDSPAmp, amp:dword


;
;Set DSP Volume
;
;This value attenuates the output and was implemented to allow songs to be faded out.  ResetDSP sets
;this value to 65536 (no attenuation).
;
;In:
;   vol = Volume [-1.16] (0.0 to 1.0, negative values act as 0)
;
;Destroys:
;   EAX

PUBLIC SetDSPVol, vol:dword
PUBLIC SetDSPLength, song:dword, fade:dword	;see SetAPULength in APU.Inc


;
;Set Voice Stereo Separation
;
;Sets the amount to adjust the panning position of each voice
;
;In:
;   sep = Separation [1.16]
;         1.0 - full separation (output is either left, center, or right)
;         0.5 - normal separation (output is unchanged)
;           0 - no separation (output is completely monaural)
;
;Destroys:
;   EAX

PUBLIC SetDSPStereo, sep:dword


;
;Set Echo Feedback Crosstalk
;
;Sets the amount of crosstalk between the left and right channel during echo feedback
;
;In:
;   leak = Crosstalk amount [-1.15]
;           1.0 - no crosstalk (SNES)
;             0 - full crosstalk (mono/center)
;          -1.0 - inverse crosstalk (L/R swapped)
;
;Destroys:
;   EAX

PUBLIC SetDSPEFBCT, leak:dword


;
;DSP Data Port
;
;Writes a value to a specified DSP register and alters the DSP accordingly.  If the register write
;affects the output generated by the DSP, this function returns true.
;
;Note:
;   SetDSPReg does not call the debugging vector
;
;In:
;   reg = DSP Address
;   val = DSP Data
;
;Out:
;   EAX = True, if the DSP state was affected

PUBLIC SetDSPReg, reg:byte, val:byte
PUBLIC DSPIn, NULL									;Used internally by SPC700.Asm


;
;Emulate DSP
;
;Emulates the DSP of the SNES
;
;Notes:
;   If 'pBuf' is null, the routine MIX_NONE will be used
;   Range checking is performed on 'len'
;
;In:
;   pBuf-> Buffer to store output
;   len  = Length of buffer (in samples)
;
;Out:
;   EAX -> End of buffer
;
;Destroys:
;   ST0-ST7

PUBLIC EmuDSP, pBuf:ptr, len:dword
PUBLIC SetEmuDSP, pBuf:ptr, len:dword, rate:dword	;Used internally by APU.Asm
PUBLIC CatchUp											;Used internally by SPC700.Asm


;
;Play Waveform
;
;Plays a single sound source using the settings defined by SetDSPOpt.  Interpolation is performed
;with floating-point math regardless of mix type.  All samples output are in integer form.
;
;Admittedly, this function is poorly coded in that it does two different things depending on the
;value of the last parameter.
;
;Note:
;   If the source data is within APU RAM, the source pointer will wrap around within APU RAM
;
;In:
;   pSrc-> Sound source to play (if upper 16 bits are 0, lower 16 bits will index APU RAM)
;   loop = Block to loop on
;   rate = Logical sample rate to play back at (actual playback rate may be less)
;
;   pSrc-> Buffer to store decompressed and resampled data
;   loop = Number of samples to generate
;   rate = 0
;
;Out:
;   rate != 0, EAX = Sample rate sound will be played back at
;   rate == 0, EAX-> End of buffer

PUBLIC PlayWave, pSrc:ptr, loop:dword, rate:dword


;
;Unpack Waveform
;
;Decompresses a series of BRR blocks
;
;Notes:
;   Regardless of the value in 'num', unpacking will stop when the end block is reached or, if the
;    source blocks are in APU RAM, when the end of APU RAM is reached.
;   The previous samples are only needed if you're making multiple calls to UnpackWave.  If you're
;    decompressing the entire sound in one call, you can disregard them.
;
;In:
;   pBuf -> Location to store decompressed samples
;   pSrc -> Sound source to decompress (if upper 16 bits are 0, lower 16 bits will index APU RAM)
;   num   = Number of blocks to decompress (0 to decompress entire sound)
;   opts  = Options (only one for now)
;           DSP_OLDSMP = Traditional method of decompression
;   prev1-> First previous sample (can be NULL if not needed)
;   prev2-> Second previous sample
;
;Out:
;   EAX = Number of blocks decompressed
;
;Destroys:
;   EAX

PUBLIC UnpackWave, pBuf:ptr, pSrc:dword, num:dword, opts:dword, prev1:ptr, prev2:ptr


;
;Pack Waveform
;
;Compresses a series of samples into BRR blocks
;
;Notes:
;   The previous samples are only needed if you're making multiple calls to PackWave.  If you're
;    compressing the entire sound in one call, you can disregard them.
;   No bounds checking is performed on the location output to make sure it resides in APU RAM.
;
;In:
;   pBuf -> Location to store compressed samples
;   pSmp -> 16-bit samples to compress
;   num   = Number of blocks to compress
;   opts  = Options
;           BRR_LINEAR = Force linear compression (filter 0)
;           BRR_LOOP = Set loop flag in block header
;   prev1-> First previous sample (can be NULL if not needed)
;   prev2-> Second previous sample
;
;Destroys:
;   EAX

PUBLIC PackWave, pBuf:ptr, pSmp:ptr, num:dword, opts:dword, prev1:ptr, prev2:ptr


;
;VMax to Decible
;
;Converts the linear values of Voice.VMaxL/R and VMMaxL/R to decible values then resets them to
;zero.  The decible values are stored in an array, where the first 16 values used for each voice and
;the last two for the main outputs.  The output values can be in fixed point integer or in floating-
;point form.
;
;Notes:
;   When floating-point numbers are chosen, the return values are in actual dB starting with -96
;    (which represents -infinity) on up to +0.
;   When integers are chosen, the return values are in fixed point notation [16.16] and are positive
;    starting with 0 (which represents -infinity) on up to +96 (or 6291456, representing +0db).
;   The values from vMMax can be higher than 0db if clipping has occurred.
;
;In:
;   pList -> Array to store 18 32-bit numbers
;   f     = True, return floating-point numbers
;           False, return integers
;
;Destroys:
;   EAX

PUBLIC VMax2dB, pList:ptr, f:byte