/***************************************************************************************************
* Program:    Super Nintendo Entertainment System(tm) Digital Signal Processor Emulator            *
* Platform:   80386 / MMX                                                                          *
* Programmer: Anti Resonance                                                                       *
*                                                                                                  *
* "SNES" and "Super Nintendo Entertainment System" are trademarks of Nintendo Co., Limited and its *
* subsidiary companies.                                                                            *
*                                                                                                  *
* This program is free software; you can redistribute it and/or modify it under the terms of the   *
* GNU General Public License as published by the Free Software Foundation; either version 2 of     *
* the License, or (at your option) any later version.                                              *
*                                                                                                  *
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;        *
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.        *
* See the GNU General Public License for more details.                                             *
*                                                                                                  *
* You should have received a copy of the GNU General Public License along with this program;       *
* if not, write to the Free Software Foundation, Inc.                                              *
* 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                        *
*                                                                                                  *
* ------------------------------------------------------------------------------------------------ *
* Revision History:                                                                                *
*                                                                                                  *
* 5.0  26.01.2004 SNESAPU v2.0                                                                     *
*      + Added 24- and 32-bit sample support to MMX                                                *
*      + Hardware anomalies emulates the DSP at 32kHz then upsamples the output                    *
*      + Replaced fake low-pass filter with a 32-tap FIR filter                                    *
*      + Writing to the DSP registers from the SPC700 automatically calls EmuDSP.  This should     *
*        help keep the DSP and SPC700 in better syncronization.                                    *
*      + Removed self modifying code                                                               *
*      + Moved fade out from APU.Asm                                                               *
*      - Output is inverted only when hardware anomalies are enabled                               *
*      - Obtained a more accurate copy of the gaussian look-up table used in the SNES              *
*      - Gaussian interpolation correctly clears the LSB in MMX mixing mode                        *
*      - When using 4-point interpolation, source playback begins with the fourth sample           *
*      - Removed 17-bit clamp from filter 1                                                        *
*      - Envelopes are not based on a 33kHz clock (duh...)                                         *
*      - Envelopes now use a sample counter instead of a fractional clock                          *
*      - FIR filter coefficients were getting applied in reverse order                             *
*      - Filtered samples get shifted right before accumulation, and LSB is cleared after accum.   *
*      - Fixed PckWav (it was seriously broken)                                                    *
*      - Added support for BRR_??? options to PckWav                                               *
*                                                                                                  *
* 4.26 29.07.2003 SNESAPU v1.01                                                                    *
*      - Voice's inactive flag wasn't getting set after a key off                                  *
*                                                                                                  *
* 4.25 30.06.2003                                                                                  *
*      - Made some minor changes to RestoreDSP                                                     *
*      - Fixed voices' pitch not getting changed in SetDSPOpt                                      *
*      - Fixed a bug in the floating-point DSP volume register handler when STEREO = 0             *
*                                                                                                  *
* 4.2  18.06.2003 SNESAPU v0.99c                                                                   *
*      - Fixed interger overflow in MMX FIR filter                                                 *
*                                                                                                  *
* 4.11 14.06.2003 SNESAPU v0.99b                                                                   *
*      + Added support for old amplification values (<= 256)                                       *
*                                                                                                  *
* 4.1  19.05.2003 SNESAPU v0.99a                                                                   *
*      - Fixed a bug in the i386 FIR filter                                                        *
*                                                                                                  *
* 4.0  28.04.2003 SNESAPU v0.99                                                                    *
*      + InitDSP will select 386 or MMX depending on CPU                                           *
*      + Added 16 and 24-bit sample output to EmuDSPF                                              *
*      + Changed SetDSPAmp to decibels                                                             *
*      - Fixed several bugs in the envelope code so envelopes now work the same as the SNES (fixes *
*        a lot of songs)                                                                           *
*      - Stopped caching loop points (fixes Castlevania IV, which likes to change sources after a  *
*        key on)                                                                                   *
*      - Added clamping to the pitch modulation code (fixes CT)                                    *
*      - Fixed a bug in KON/KOF emulation                                                          *
*      - Changed 16-bit mixing process to emulate the SNES                                         *
*      - Improved low-pass filtering and simulated RF interference                                 *
*      - Removed floating-point support from SetDSPAmp                                             *
*      - Fixed another volume bug in SetDSPOpt when switching between integer and float routines   *
*                                                                                                  *
* 3.0  10.02.2003 SNESAPU v0.98                                                                    *
*      + Converted to NASM                                                                         *
*      + Added SetDSPDbg, SaveDSP, RestoreDSP, and SetDSPReg                                       *
*      + Added a loop to EmuDSP so the number of samples passed in can be any value                *
*      + Added floating-point support to SetDSPAmp                                                 *
*      + Added micro ramping to the channel volumes in the floating-point routine                  *
*      + Added 32-bit IEEE 754 output                                                              *
*      - Added a 17-bit clamp to the UnpckWav (fixes Hiouden; CT and FF6 are iffy now)             *
*      - Added linear interpolation to FIR filter (fixes some songs that turned to noise)          *
*      - SetDSPOpt wasn't clearing EAX, which was causing dspDecmp to sometimes contain a bad value*
*      - ENVX always gets updated, regardless of ADSR/GAIN flag (fixes Clue)                       *
*                                                                                                  *
* 2.1  xx.06.2001 SNESAPU v0.96                                                                    *
*      - freqTab was built from the wrong values (Hz instead of samples)                           *
*      - rateTab now uses 32-bits of precision instead of 16                                       *
*      - Values in rateTab >= 1.0 are now set to 0.99 to reduce the ill effects of fast envelopes  *
*        when the sample rate is < 32kHz                                                           *
*      - Reset all internal pitch values when PMON is written to (fixes Asterix)                   *
*      - Clear bits in KOF if voice is inactive (improves EOS detection)                           *
*      - Switched to a static lookup table generated by the SNES for Gaussian interpolation        *
*      - Damn near cracked the sample decompression                                                *
*      - Optimized code a bit in EmuDSPN and inadvertently fixed a bug that was causing some songs *
*        to break after seeking                                                                    *
*      - EmuDSPN wasn't checking the loop counter after an envelope update was performed (seeking  *
*        no longer appears to hang)                                                                *
*                                                                                                  *
* 2.0  15.05.2001 SNESAPU v0.95                                                                    *
*      + Added 4-point Gaussian interpolation                                                      *
*      + EmuDSPN now decompresses samples, so seeking doesn't lose voices                          *
*      + Added an option to FixSeek to disable resetting voices                                    *
*      + Reverse stereo swaps the left/right volumes instead of physically swapping samples        *
*      + More minor optimizations                                                                  *
*      - Changed the sample decompression routine.  Less songs should have problems.               *
*      - Echo delay was incorrect for most sample rates                                            *
*      - Did something to fix the noise generator not working in some songs                        *
*                                                                                                  *
* 1.9  02.03.2001 SNESAPU v0.92                                                                    *
*      - Fixed the low-pass filter (I'm amazed it was sort of working at all)                      *
*      - Added a hack to UnpckWav to subtract 3 from the range if > 12                             *
*      - Forgot that the DSP needs to be limited to 64K as well (Secret of Evermore doesn't crash) *
*      - Fixed a problem in the MMX dynamic code allocation                                        *
*      - FixSeek wasn't resetting the DSP X registers (seeking works in Der Langrisser)            *
*                                                                                                  *
* 1.8  26.01.2001 SNESAPU v0.90                                                                    *
*      + Added voice monitor and VMax2dB for visualization                                         *
*      + Added SetDSPPitch so people with SB/SBPro's and compatibles can hear the correct tone     *
*      + Added SetDSPStereo for headphone listeners                                                *
*      + Added SetDSPEFBCT to make the echo sound more natural                                     *
*      + Added UnpackWave so waveforms could be decompressed by external programs                  *
*      + Added 32-bit floating point mixer for people with 24/32-bit sound cards                   *
*      + Added a reverse stereo option                                                             *
*      + Added GetProcInfo to detect 3DNow! and SSE capabilites                                    *
*      + Added EmuDSP which automatically calls the mixing routine set in SetDSPOpt                *
*      + Made a lot of minor optimizations to DSPIn and EmuDSP                                     *
*      + Made separate register handlers in DSPIn for mono and floating-point                      *
*      - Changed SetDSPOpt to use individual parameters instead of a record                        *
*      - SetDSPOpt performs range checking                                                         *
*      - Fixed configuration implementation so options don't need to be reset for every song and   *
*        they can be changed between calls to EmuDSP                                               *
*      - Fixed mono related code (sounds better now)                                               *
*      - Echo filter was never getting turned on                                                   *
*      - Implemented a better low-pass filter                                                      *
*                                                                                                  *
* 1.4  22.05.2000 SNESAmp v1.2                                                                     *
*      + Added a low-pass filter                                                                   *
*      + Added output monitor                                                                      *
*      + Optimized the waveform code in EmuDSPN                                                    *
*      + Added FixSeek                                                                             *
*      + SetDSPVol now uses a 16-bit value                                                         *
*      - FixDSP was marking voices as keyed even if the pitch was 0                                *
*      - Correctly implemented echo volume (should be perfect now)                                 *
*                                                                                                  *
* 1.1  05.04.2000 SNESAmp v1.0                                                                     *
*      - Voices with a pitch of 0 can't be keyed on                                                *
*      - EnvH becomes 0 when voice is keyed off                                                    *
*      - EnvH is set to 0 when envelope is in gain mode                                            *
*      - Account for sample blocks with a range greater than 16-bits (ADPCM still not correct)     *
*                                                                                                  *
* 1.0  17.03.2000 SNESAmp v0.9                                                                     *
*                                                      Copyright (C)1999-2004 Alpha-II Productions *
***************************************************************************************************/

#ifndef	__INC_DSP
#define	__INC_DSP

//**************************************************************************************************
//Defines

//CPU capability (for GetProcType) -------------
#define	CPU_MMX		1							//Multi-Media eXtensions
#define	CPU_3DNOW	2							//3DNow! extensions
#define	CPU_SSE		4							//Streaming SIMD Extensions

//Mixing routines (see DSP.ASM for details of each routine)
#define	MIX_NONE	0							//No mixing
#define	MIX_INT		1							//Use integer math
#define	MIX_FLOAT	3							//Use floating-point math

//Interpolation routines -----------------------
#define	INT_NONE	0							//None
#define	INT_LINEAR	1							//Linear
#define	INT_CUBIC	2							//Cubic
#define	INT_GAUSS	3							//Gaussian

//DSP options ----------------------------------
#define	DSP_ANALOG	0x01						//Simulate analog anomalies
#define	DSP_OLDSMP	0x02						//Old ADPCM sample decompression routine
#define	DSP_SURND	0x04						//"Surround" sound
#define	DSP_REVERSE	0x08						//Reverse stereo samples
#define	DSP_NOECHO	0x10						//Disable echo

//PackWave options -----------------------------
#define	BRR_LINEAR	0x01						//Use linear compression for all blocks
#define	BRR_LOOP	0x02						//Set loop flag in block header
#define	BRR_NOINIT	0x04						//Don't create an initial block of silence
#define	BRR_8BIT	0x10						//Input samples are 8-bit


//**************************************************************************************************
//Public Variables

//DSP registers --------------------------------
typedef struct DSPVoice
{
	s8	volL;									//Volume Left
	s8	volR;									//Volume Right
	u16	pitch;									//Pitch (rate/32000) (3.11)
	u8	srcn;									//Sound source being played back
	u8	adsr[2];								//Envelope rates for attack, decay, and sustain
	u8	gain;									//Envelope gain (if not using ADSR)
	s8	envx;									//Current envelope height (.7)
	s8	outx;									//Current sample being output (-.7)
	s8	__r[6];
} DSPVoice;

typedef struct DSPFIR
{
	s8	__r[15];
	s8	c;										//Filter coefficient
} DSPFIR;

typedef union DSPReg
{
	DSPVoice	voice[8];						//Voice registers

	struct										//Global registers
	{
		s8	__r00[12];
		s8	mvolL;								//Main Volume Left (-.7)
		s8	efb;								//Echo Feedback (-.7)
		s8	__r0E;
		s8	c0;									//FIR filter coefficent (-.7)

		s8	__r10[12];
		s8	mvolR;								//Main Volume Right (-.7)
		s8	__r1D;
		s8	__r1E;
		s8	c1;

		s8	__r20[12];
		s8	evolL;								//Echo Volume Left (-.7)
		u8	pmon;								//Pitch Modulation on/off for each voice
		s8	__r2E;
		s8	c2;

		s8	__r30[12];
		s8	evolR;								//Echo Volume Right (-.7)
		u8	non;								//Noise output on/off for each voice
		s8	__r3E;
		s8	c3;

		s8	__r40[12];
		u8	kon;								//Key On for each voice
		u8	eon;								//Echo on/off for each voice
		s8	__r4E;
		s8	c4;

		s8	__r50[12];
		u8	kof;								//Key Off for each voice (instantiates release mode)
		u8	dir;								//Page containing source directory (wave table offsets)
		s8	__r5E;
		s8	c5;

		s8	__r60[12];
		u8	flg;								//DSP flags and noise frequency
		u8	esa;								//Starting page used to store echo waveform
		s8	__r6E;
		s8	c6;

		s8	__r70[12];
		u8	endx;								//Waveform has ended
		u8	edl;								//Echo Delay in ms >> 4
		s8	__r7E;
		s8	c7;
	};

	DSPFIR	fir[8];								//FIR filter

	u8		reg[128];
} DSPReg;

//Internal mixing data -------------------------
typedef struct MixF
{
	b8	mute:1;									//Voice is muted (set by user)
	u8	__r1:1;
	b8	keyOff:1;								//Voice is in key off mode
	b8	inactive:1;								//Voice is inactive, no samples are being played
	u8	__r2:2;
	u8	intType:2;								//Interpolation type (no longer used)
} MixF;

typedef enum EnvM
{
	ENV_DEC,									//Linear decrease
	ENV_EXP,									//Exponential decrease
	ENV_INC,									//Linear increase
	ENV_BENT = 6,								//Bent line increase
	ENV_DIR,									//Direct setting
	ENV_REL,									//Release mode (key off)
	ENV_SUST,									//Sustain mode
	ENV_ATTACK,									//Attack mode
	ENV_DECAY = 13,								//Decay mode
} EnvM;

#define	ENVM_IDLE	0x80						//Envelope is marked as idle, or not changing
#define	ENVM_MODE	0xF							//Envelope mode is stored in lower four bits

typedef struct Voice
{
	//Voice -----------08
	DSPVoice	*pVoice;						//-> voice registers in DSP
	u32		_r;
	//Waveform --------06
	void	*bCur;								//-> current block
	u8		bHdr;								//Block Header for current block
	u8		mFlg;								//Mixing flags (see MixF)
	//Envelope --------22
	u8		eMode;								//[3-0] Current mode (see EnvM)
												//[6-4] ADSR mode to switch into from Gain
												//[7]   Envelope is idle
	u8		eRIdx;								//Index in RateTab (0-31)
	u32		eRate;								//Rate of envelope adjustment (16.16)
	u32		eCnt;								//Sample counter (16.16)
	u32		eVal;								//Current envelope value
	s32		eAdj;								//Amount to adjust envelope height
	u32		eDest;								//Envelope Destination
	//Visualization ---08
	s32		vMaxL;								//Maximum absolute sample output
	s32		vMaxR;
	//Samples ---------52
	s32		sP1;								//Last sample decompressed (prev1)
	s32		sP2;								//Second to last sample (prev2)
	s16		*sIdx;								//-> current sample in sBuf
	s16		sBufP[4];							//Last 4 samples from previous block (needed for inter.)
	s16		sBuf[16];							//32 bytes for decompressed sample blocks
	//Mixing ----------32
	f32		mTgtL;								//Target volume (floating-point routine only)
	f32		mTgtR;								// "  "
	s32		mChnL;								//Channel Volume (-24.7)
	s32		mChnR;								// "  "
	u32		mRate;								//Pitch Rate after modulation (16.16)
	u32		mDec;								//Pitch Decimal (.16) (used as delta for interpolation)
	u32		mOrgP;								//Original pitch rate converted from the DSP (16.16)
	s32		mOut;								//Last sample output before chn vol (used for pitch mod)
} Voice;

//Saved state ----------------------------------
//see SaveDSP for information about the values in the structure, and how they differ from the
//internal values
typedef struct DSPState
{
	DSPReg	*pDSP;								//[0.0] -> DSP registers (128 bytes)
	Voice	*pVoice;							//[0.4] -> Internal mixing settings (1k)
	void	*pEcho;								//[0.8] -> echo buffer (bytes = sample rate * 1.92)
	u32		_r1;								//[0.C] reserved

	u32		vMMaxL,vMMaxR;						//[1.0] Maximum output so far
	u32		mAmp;								//[1.8] Amplification
	u8		vActive;							//[1.C] Flags for each active voice
	u8		_r2[3];								//[1.D] reserved
} DSPState;


//**************************************************************************************************
// DSP Debugging Routine
//
// A prototype for a function that gets called for debugging purposes.
//
// The paramaters passed in can be modified, and on return will be used to update the internal
// registers.  Set bit-7 of 'reg' to prevent the DSP from handling the write.
//
// For calling a pointer to another debugging routine, use the following macro.
//
// In:
//    reg -> Current register (LOBYTE = register index)
//    val  = Value being written to register

typedef void DSPDebug(volatile u8 *reg, volatile u8 val);

#ifdef	__GNUC__
#define	_CallDSPDebug(proc, reg, val) \
			asm(" \
				pushl	%2 \
				pushl	%1 \
				calll	%0 \
				popl	%1 \
				popl	%2 \
			" : : "m" (proc), "m" (reg), "m" (val));
#else
#define	_CallDSPDebug(proc, reg, val) \
			_asm \
			{ \
				push	dword ptr [val]; \
				push	dword ptr [reg]; \
				call	dword ptr [proc]; \
				pop		dword ptr [reg]; \
				pop		dword ptr [val]; \
			}
#endif


////////////////////////////////////////////////////////////////////////////////////////////////////
// Private Declarations

#ifndef	SNESAPU_DLL

#ifdef	__cplusplus
extern "C"	DSPReg	dsp;						//DSP registers
extern "C"	Voice	mix[8];						//Mixing structures for each voice
extern "C"	u32		vMMaxL,vMMaxR;				//Maximum absolute sample output
#else
extern DSPReg	dsp;
extern Voice	mix[8];
extern u32		vMMaxL,vMMaxR;
#endif


////////////////////////////////////////////////////////////////////////////////////////////////////
// External Functions
//
// NOTE:  Unless otherwise stated, no range checking is performed.  It is the caller's
//        responsibilty to ensure paramaters passed are within a valid range.

#ifdef	__cplusplus
extern "C" {
#endif

//**************************************************************************************************
// Get Processor Information
//
// Detects if processor has MMX, 3DNow!, or SSE capabilities
//
// Note:
//    Values are or'd together.  If a procesor has 3DNow!, it will return 03h for MMX and 3DNow!.
//
// Out:
//    see CPU_xxx #defines
//
// Destroys:
//    nothing

u32 GetProcInfo();


//**************************************************************************************************
// Initialize DSP
//
// Creates the lookup tables for interpolation, and sets the default mixing settings:
//
//    mixType = MIX_INT
//    numChn  = 2
//    bits    = 16
//    rate    = 32000
//    inter   = INT_GAUSS
//    opts    = 0
//
// Note:
//    Callers should use InitAPU instead
//
// In:
//    -> 64KB of APU RAM aligned on a 64K boundary
//
// Destroys:
//    ST0-7

void InitDSP(void*);


//**************************************************************************************************
// Reset DSP
//
// Resets the DSP registers, erases internal variables, and resets the volume
//
// Note:
//    Callers should use ResetAPU instead
//
// Destroys:
//    EAX

void ResetDSP();


//**************************************************************************************************
// Set DSP Options
//
// Recalculates tables, changes the output sample rate, and sets up the mixing routine
//
// Notes:
//    Range checking is performed on all parameters.  If a parameter does not match the required
//     range of values, the default value will be assumed.
//
//    -1 can be used for any paramater that should remain unchanged.
//
//    The DSP_LOW option only works when:
//     mix  =  MIX_INT
//     chn  =  2
//     bits >= 16
//     rate >= 32000
//
//    Callers should use SetAPUOpt instead
//
// In:
//    mix   = Mixing routine (default MIX_INT)
//    chn   = Number of channels (1 or 2, default 2)
//    bits  = Sample size (mono 8 or 16; stereo 8, 16, 24, 32, or -32 [IEEE 754], default 16)
//    rate  = Sample rate (8000-192000, default 32000)
//    inter = Interpolation type (default INT_GAUSS)
//    opts  = See 'DSP options' in the Defines section
//
// Destroys:
//    EAX

void SetDSPOpt(u32 mix, u32 chn, u32 bits, u32 rate, u32 inter, u32 opts);


//**************************************************************************************************
// Debug DSP
//
// Installs a vector that gets called each time a value is written to the DSP data register.
//
// Note:
//    The build option SA_DEBUG must be enabled
//
// In:
//    pTrace-> debug function (a null pointer turns off the debug call)
//
// Out:
//    Previously installed vector

DSPDebug* SetDSPDbg(DSPDebug *pTrace);


//**************************************************************************************************
// Fix DSP After Loading SPC File
//
// Initializes the internal mixer variables
//
// Note:
//    Callers should use FixAPU instead
//
// Destroys:
//    EAX

void FixDSP();


//**************************************************************************************************
// Fix DSP After Seeking
//
// Puts all DSP voices in a key off state and erases echo region.  Call after using EmuDSPN.
//
// In:
//    True  = Reset all voices
//    False = Only erase memory
//
// Destroys:
//    EAX

void FixSeek(u8 reset);


//**************************************************************************************************
// Save DSP State
//
// Notes:
//    Pointers in the DSPState structure can be NULL
//
//    Pointers in the Voice structure are not returned as literal values, but as an index instead:
//     Voice.pVoice = index into APU RAM (0-65535)
//     Voice.bCur   = index into APU RAM (0-65535)
//     Voice.sIdx   = index into sBuf (0-15)
//
//    Callers should use SaveAPU instead
//
// In:
//    pState -> Saved state structure
//
// Destroys:
//    EAX

void SaveDSP(DSPState *state);


//**************************************************************************************************
// Restore DSP State
//
// Notes:
//    Setting a pointer in DSPState to NULL will keep its corresponding internal data intact, except
//     for the echo.  If pEchoBuf is NULL, the internal echo buffer will be cleared.
//
//    Pointers in the Voice structure must contain index values, not literal pointers (see SaveDSP)
//
//    Because the envelope (e???) and mixing (m???) variables contain values affected by the current
//     sample rate and size, these values may need to be reset.  By setting eRate or mRate to -1,
//     RestoreDSP will recaluclate their respective values.
//
//    When setting eRate to -1, eMode must contain the envelope mode when the state was saved, and
//     eVal must contain the envelope height as a 7-bit quantity (as in ENVX), not a fixed point
//     quantity as it was saved out.
//
//    mFlg is not affected by setting mRate to -1.
//
// In:
//    pState -> Saved state structure
//
// Destroys:
//    EAX

void RestoreDSP(DSPState *pState);


//**************************************************************************************************
// Set DSP Base Pitch
//
// Adjusts the pitch of the DSP
//
// In:
//    base = Base sample rate (32000 = Normal pitch, 32458 = Old SB cards, 32768 = Old ZSNES)
//
// Destroys:
//    EAX

void SetDSPPitch(u32 base);


//**************************************************************************************************
// Set DSP Amplification
//
// This value is applied to the output with the main volumes
//
// In:
//    amp = Amplification level [-15.16] (1.0 = SNES, negative values act as 0)
//
// Destroys:
//    EAX

void SetDSPAmp(u32 amp);


//**************************************************************************************************
// Set DSP Volume
//
// This value attenuates the output and was implemented to allow songs to be faded out.
//
// Notes:
//    ResetDSP sets this value to 65536 (no attenuation).
//    This function is called internally by EmuAPU and SetAPULength, and should not be called by
//     the user.
//
// In:
//    vol = Volume [-1.16] (0.0 to 1.0, negative values act as 0)
//
// Destroys:
//    EAX

void SetDSPVol(u32 vol);


//**************************************************************************************************
// Set Voice Stereo Separation
//
// Sets the amount to adjust the panning position of each voice
//
// In:
//   sep = Separation [1.16]
//         1.0 - full separation (output is either left, center, or right)
//         0.5 - normal separation (output is unchanged)
//           0 - no separation (output is completely monaural)
//
// Destroys:
//    EAX,ST(0-7)

void SetDSPStereo(u32 sep);


//**************************************************************************************************
// Set Echo Feedback Crosstalk
//
// Sets the amount of crosstalk between the left and right channel during echo feedback
//
// In:
//   leak = Crosstalk amount [-1.15]
//           1.0 - no crosstalk (SNES)
//             0 - full crosstalk (mono/center)
//          -1.0 - inverse crosstalk (L/R swapped)
//
// Destroys:
//    EAX

void SetDSPEFBCT(s32 leak);


//**************************************************************************************************
// DSP Data Port
//
// Writes a value to a specified DSP register and alters the DSP accordingly.  If the register write
// affects the output generated by the DSP, this function returns true.
//
// Note:
//    SetDSPReg does not call the debugging vector
//
// In:
//    reg = DSP Address
//    val = DSP Data
//
// Out:
//    true, if the DSP state was affected

b8 SetDSPReg(u8 reg, u8 val);


//**************************************************************************************************
// Emulate DSP
//
// Emulates the DSP of the SNES
//
// Notes:
//    If 'pBuf' is NULL, the routine MIX_NONE will be used
//    Range checking is performed on 'size'
//
//    Callers should use EmuAPU instead
//
// In:
//    pBuf-> Buffer to store output
//	  len  = Length of buffer (in samples, can be 0)
//
// Out:
//    -> End of buffer
//
// Destroys:
//    ST(0-7)

void* EmuDSP(void *pBuf, s32 size);


//**************************************************************************************************
// Play Waveform
//
// Plays a single sound source using the settings defined by SetDSPOpt.  Interpolation is performed
// with floating-point math regardless of mix type.  All samples output are in integer form.
//
// Admittedly, this function is poorly coded in that it does two different things depending on the
// value of the last parameter.
//
// In:
//    pSrc-> Sound source in APU RAM to play (only the lower 16-bits are used)
//    loop = Block to loop on
//    rate = Logical sample rate to play back at (actual playback rate may be less)
//
//    pSrc-> Buffer to store decompressed and resampled data
//    loop = Number of samples to generate
//    rate = 0
//
// Out:
//    rate != 0, return value is sample rate sound will be played back at
//    rate == 0, return value -> end of buffer

void* PlayWave(void *pSrc, u32 loop, u32 rate);


//**************************************************************************************************
// Unpack Waveform
//
// Decompresses a series of BRR blocks
//
// Notes:
//    Unpacking will stop when the end block or the end of APU RAM is reached, regardless of the
//    value in num.
//    The previous samples are only needed if you want to stop and resume a decompression stream.
//     If you're decompressing the entire sound in one call, you can disregard them.
//
// In:
//    pBuf -> Location to store decompressed samples
//    pSrc -> Sound source in APU RAM to decompress (only the lower 16-bits are used)
//    num   = Number of blocks to decompress (0 to decompress entire sound)
//    opts  = Options (only one for now)
//            DSP_OLDSMP = Traditional method of decompression
//    prev1-> First previous sample (can be NULL)
//    prev2-> Second previous sample
//
// Out:
//    Number of blocks decompressed
//
// Destroys:
//    EAX

u32 UnpackWave(s16 *pBuf, void *pSrc, u32 num, u32 opts, s32 *prev1, s32 *prev2);


//**************************************************************************************************
// Pack Waveform
//
// Compresses a series of samples into BRR blocks
//
// Notes:
//    The previous samples are only needed if you want to compress samples a block at a time.
//    pBuf can point anywhere, not necessarily somewhere within APU RAM.  No bounds checking is
//     performed.
//
// In:
//    pBuf -> Location to store compressed samples
//    pSmp -> 16-bit samples to compress
//    num   = Number of blocks to compress
//    opts  = Options
//            BRR_LINEAR = Force linear compression (filter 0)
//            BRR_LOOP = Set loop flag in block header
//    prev1-> First previous sample (can be NULL if not needed)
//    prev2-> Second previous sample
//
// Destroys:
//    EAX

void PackWave(void *pBuf, s16 *pSmp, u32 num, u32 opts, s32 *prev1, s32 *prev2);


//**************************************************************************************************
// VMax to Decible
//
// Converts the linear values of Voice.VMaxL/R and VMMaxL/R to decible values then resets them to
// zero.  The decible values are stored in an array, where the first 16 values used for each voice
// and the last two for the main outputs.  The output values can be in fixed point integer or in
// floating-point form.
//
// Notes:
//    When floating-point numbers are chosen, the return values are in actual dB starting with -96
//     (which represents -infinity) on up to +0.
//    When integers are chosen, the return values are in fixed point notation [16.16] and are
//     positive starting with 0 (which represents -infinity) on up to +96 (or 6291456, representing
//      +0db).
//    The values from vMMax can be higher than 0db if clipping has occurred.
//
// In:
//    pList-> Array to store 18 32-bit numbers
//    fp    = True, return floating-point numbers
//            False, return integers
//
// Destroys:
//    EAX

void VMax2dB(void *pList, b8 fp);

#ifdef	__cplusplus
}
#endif

#endif	//SNESAPU_DLL
#endif	//__INC_DSP
