;
;Program:    Sony SPC700 Emulator
;Platform:   80386
;Programmer: Anti Resonance
;
;Thanks to Michael Abrash.  It was reading the Graphics Programming Black Book that gave me the
;idea and inspiration to write this in the first place.
;
; This program is free software; you can redistribute it and/or modify it under the terms of the
; GNU General Public License as published by the Free Software Foundation; either version 2 of
; the License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
; See the GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License along with this program;
; if not, write to the Free Software Foundation, Inc.
; 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;
;Revision History:
;
; 2.4   26.01.2004 SNESAPU v2.0
;       - Added code to call DSP emulator if 00F3 is read from
;       - Updated opcode fetch code to work with new DSP syncronization
;
; 2.31  30.06.2003
;       - Fixed a bug where the counter would increase every pulse if the timer was disabled and set
;         to 0
;
; 2.3   10.02.2003 SNESAPU v0.98
;       + Added SaveSPC, RestoreSPC, and SetAPURAM
;       + Coded support for BRK (still not sure if it's correct)
;       - Rewrote SetSPCDbg and the opcode fetcher to fix some potential bugs
;       - Added code to CntHack to try and isolate small polling loops (fixes the tempo problems
;         some games were having)
;       - Changed FixSPC, again, to copy extraRAM from APURAM[0FFC0h] if IPL reading is disabled in
;         F1h (fixes Tales of Phantasia)
;       - Fixed a lame bug when clearing the in-ports with F1h (fixes Dragon Quest)
;
; 2.25  12.04.2001 SNESAPU v0.95
;       + Added SPC_NODSP to the debugging options
;       - Fixed some bugs with copying the IPL ROM and detecting reading ability
;
; 2.21  29.01.2001 SNESAPU v0.92
;       - Changed internal clock to 1.024MHz (I haven't noticed any differences, yet)
;
; 2.2   01.01.2001 SNESAPU v0.90
;       + Added an external counter to the 64kHz timer
;       + Emulator is now based off of a 24.576MHz clock, so implementation of a 2.048 or 2.457MHz
;         CPU clock will be easier internally and transparent externally
;       + Added speed hack to SLEEP
;       + Added option flags to DebugSPC to allow greater control over the SPC700 state
;       + Added breaking ability to STOP in non-debug builds
;       - In ports don't get cleared on load, regardless of setting in F1
;       - Speed hack is now implemented properly
;       - PC is now operated on as a 16-bit quantity (Rudra no Hihou expects PC to roll over,
;         which it wasn't doing when instructions were adding to ESI instead of SI)
;       - Changed coding style a bit to make code easier to read and added more comments
;
; 2.11  04.05.2000 SNESAmp v1.2
;       - Control register is now set to 80h on reset
;
; 2.1   04.04.2000 SNESAmp v1.0
;       + Rearranged code so static branch prediction would be more accurate
;       + Clock cycles to emulate is passed on the stack to EmuSPC
;       + SPC700 registers are passed on the stack to the external debug routine
;       + Eliminated ORG directive and rewrote code to conform to MASM syntax
;       - Fixed some implementation bugs with the speed hack
;       - Fixed a bug in the 16-bit WrPost macro, again
;
; 2.0   17.03.2000 SNESAmp v0.9
;       + Rewrote for compatibility in a 32-bit C environment
;       + Added external functions to write to ports
;       - SLEEP wasn't erasing the port status on the first time through
;
; 1.51  22.02.2000 SPCTool v0.6
;       - Fixed a bug in the 16-bit WrPost macro
;
; 1.5   01.02.2000 SPCTool v0.51
;       + Added a hack to eliminate polling the counters
;       + Aligned many of the short jump destinations on paragraph boundaries
;       + Rewrote the opcode header macro to compile into table form instead of packed form.
;         (The compiled code is four times bigger, but it's now easier to align, port, and debug.)
;       + Store flags as dwords so accesses to the DP don't cause a partial register stall on PII's
;       - Fixed PCALL, STOP, and SLEEP
;
; 1.1   21.01.2000 SPCTool v0.5
;	     + Reprogrammed PSW handling for faster entry/exit
;
; 1.0   27.12.1999 SPCTool v0.2
;                                                        Copyright (C)1999-2004 Alpha-II Productions
;

%define SPC700_INC

;
;Equates

APU_CLK		EQU	24576000

;SPC700 debugging options -------------------
SPC_RETURN	EQU	1									;Return from EmuSPC
SPC_HALT		EQU	2									;Halt emulation
DSP_HALT		EQU	4									;Internal flag used for syncing SPC_HALT with EmuDSP
SPC_NODSP	EQU	8									;Disable calls to the DSP emu from the SPC700
SPC_TRACE	EQU	10h								;Trace next instruction


;
;Structures

STRUC SPCState
	pRAM		resd	1									;[0.0] -> APU's RAM (64k)
	_rs1		resd	1
	rPC		resw	1									;[0.8] Registers
	rA			resb	1
	rY			resb	1
	rX			resb	1
	rPSW		resb	1
	rSP		resw	1

	tmr8kHz	resd	1									;[1.0] # of cycles left until timer increase
	tmr64kHz	resd	1
	tmr64Cnt	resd	1									;[1.8] # of 64kHz ticks since emulation began
	_rs2		resd	1									;[1.C] reserved
	upCnt		resb	3									;[2.0] Up counters for counter increase
	inPMod	resb	1									;[2.3] Flags for in ports that have been modified
	outP		resb	4									;[2.4] Out ports
	_rs3		resd	2									;[2.8] reserved
ENDSTRUC


;
;Public Variables

PUBLIC	extraRAM										;RAM used for storage if ROM reading is enabled
PUBLIC	outPort										;Four out ports
PUBLIC	t64Cnt										;Counter increased every 64kHz
PUBLIC	dbgOpt										;SPC700 debugging options


;
;Exported Functions

;
;Initialize SPC700
;
;This function is a remnant from the 16-bit assembly when dynamic code reallocation was used.
;Now it just initializes internal pointers.
;
;In:
;   pAPURAM -> 64KB of SPC RAM aligned on a 64K boundary
;
;Destroys:
;   EAX

PUBLIC InitSPC, pAPURAM:ptr


;
;Reset SPC700
;
;Clears all memory, resets the function registers, T64Cnt, and halt flag, and copies ROM into the
;IPL region.
;
;Destroys:
;   EAX=0

PUBLIC ResetSPC, NULL


;
;Debug SPC700
;
;Installs a vector that gets called between each opcode for debugging purposes.  The function is
;called with the C calling convention.
;
;Upon entrance to the function:
;   [ESP+4] -> Current opcode (low word = PC)
;   [ESP+8]  = YA
;   [ESP+12] = X
;   [ESP+16] = PSW
;   [ESP+20]-> Current stack (low word = SP)
;   [ESP+24] = Clock cycle down count
;              [0-1] 8kHz cycles left until counters 0 and 1 increase
;              [2]   64kHz cycles left until counter 2 increases
;              [3]   CPU cycles left until 64kHz clock pulse
;
;Note:
;   pTrace is always called when a STOP instruction is encountered, regardless of the options.
;   DEBUG must be set to 1 in APU.Inc for pTrace to be called under other circumstances
;
;In:
;   pTrace-> debug function (NULL turns off the debug call, -1 leaves the current vector)
;   opts   = SPC700 debugging options (-1 leaves the current options)
;
;Out:
;   Previously installed vector

PUBLIC SetSPCDbg, pTrace:ptr, opts:dword


;
;Fix SPC700 After Load
;
;Loads timer steps with the values in the timer registers, resets the counters, sets up the in/out
;ports, and stores the registers.
;
;In:
;   SPC internal registers
;
;Destroys:
;   EAX

PUBLIC FixSPC, pc:word, a:byte, y:byte, x:byte, psw:byte, sp:byte


;
;Get SPC700 Registers
;
;Returns the internal registers of the SPC700
;
;In:
;   Pointers to variables to store registers
;
;Destroys:
;   EAX

PUBLIC GetSPCRegs, pPC:ptr, pA:ptr, pY:ptr, pX:ptr, pPSW:ptr, pSP:ptr


;
;Save SPC700 State
;
;Note:
;   Pointers in the SPCState structure can be NULL
;
;In:
;   pState -> Saved state structure
;
;Destroys:
;   EAX

PUBLIC SaveSPC, pState:ptr


;
;Restore SPC700 State
;
;Note:
;   Setting a pointer in SPCState to NULL will keep its corresponding internal data intact.
;
;In:
;   pState -> Saved state structure
;
;Destroys:
;   EAX

PUBLIC RestoreSPC, pState:ptr


;
;Write to APU RAM
;
;Writes a value to APU RAM.  Use this instead of writing to RAM directly so any necessary internal
;changes can be made.
;
;In:
;   addr = Address to write to (only the lower 16-bits are used)
;   val  = Value to write
;
;Destroys:
;   EAX

PUBLIC SetAPURAM, addr:dword, val:byte


;
;Write to SPC700 Port(s)
;
;Writes a value to the SPC700 via the in ports.  This is provided as an alternative to APUIn.
;PortIn writes a single byte value while PortInW writes a 16-bit value to two ports.
;
;In:
;   port = Port on which to write (only the lower 2-bits are used)
;   val  = Value to write
;
;Destroys:
;   EAX

PUBLIC InPort, addr:dword, val:byte
PUBLIC InPortW, addr:dword, val:word


;
;Emulate SPC700
;
;Emulates the SPC700 for the number of clock cycles specified, or if the counter break option is
;enabled, until a counter is increased, whichever happens first.
;
;In:
;   cyc = Number of 24.576MHz clock cycles to execute (signed, must be > 0)
;
;Out:
;   EAX = Clock cycles left to execute (negative if more cycles than specified were emulated)
;
;Destroys:
;   nothing

PUBLIC EmuSPC, cyc:dword