/***************************************************************************************************
* Program:    Sony SPC700 Emulator                                                                 *
* Platform:   80386                                                                                *
* Programmer: Anti Resonance                                                                       *
*                                                                                                  *
* Thanks to Michael Abrash.  It was reading the Graphics Programming Black Book that gave me the   *
* idea and inspiration to write this in the first place.                                           *
*                                                                                                  *
* This program is free software; you can redistribute it and/or modify it under the terms of the   *
* GNU General Public License as published by the Free Software Foundation; either version 2 of     *
* the License, or (at your option) any later version.                                              *
*                                                                                                  *
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;        *
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.        *
* See the GNU General Public License for more details.                                             *
*                                                                                                  *
* You should have received a copy of the GNU General Public License along with this program;       *
* if not, write to the Free Software Foundation, Inc.                                              *
* 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                        *
*                                                                                                  *
* ------------------------------------------------------------------------------------------------ *
* Revision History:                                                                                *
*                                                                                                  *
* 2.4   26.01.2004 SNESAPU v2.0                                                                    *
*       - Added code to call DSP emulator if 00F3 is read from                                     *
*       - Updated opcode fetch code to work with new DSP syncronization                            *
*                                                                                                  *
* 2.31  30.06.2003                                                                                 *
*       - Fixed a bug where the counter would increase every pulse if the timer was disabled and   *
*         set to 0                                                                                 *
*                                                                                                  *
* 2.3   10.02.2003 SNESAPU v0.98                                                                   *
*       + Added SaveSPC, RestoreSPC, and SetAPURAM                                                 *
*       + Coded support for BRK (still not sure if it's correct)                                   *
*       - Rewrote SetSPCDbg and the opcode fetcher to fix some potential bugs                      *
*       - Added code to CntHack to try and isolate small polling loops (fixes the tempo problems   *
*         some games were having)                                                                  *
*       - Changed FixSPC, again, to copy extraRAM from APURAM[0FFC0h] if IPL reading is disabled   *
*         in F1h (fixes Tales of Phantasia)                                                        *
*       - Fixed a lame bug when clearing the in-ports with F1h (fixes Dragon Quest)                *
*                                                                                                  *
* 2.25  12.04.2001 SNESAPU v0.95                                                                   *
*       + Added SPC_NODSP to the debugging options                                                 *
*       - Fixed some bugs with copying the IPL ROM and detecting reading ability                   *
*                                                                                                  *
* 2.21  29.01.2001 SNESAmp v2.1                                                                    *
*       - Changed internal clock to 1.024MHz (I haven't noticed any differences, yet)              *
*                                                                                                  *
* 2.2   01.01.2001 SNESAmp v2.0                                                                    *
*       + Added an external counter to the 64kHz timer                                             *
*       + Emulator is now based off of a 24.576MHz clock, so implementation of a 2.048 or 2.457MHz *
*         CPU clock will be easier internally and transparent externally                           *
*       + Added speed hack to SLEEP                                                                *
*       + Added option flags to SPCDebug to allow greater control over the SPC700 state            *
*       + Added breaking ability to STOP in non-debug builds                                       *
*       - In ports don't get cleared on load, regardless of setting in F1                          *
*       - Speed hack is now implemented properly                                                   *
*       - PC is now operated on as a 16-bit quantity (Rudra no Hihou expects PC to roll over,      *
*         which it wasn't doing when instructions were adding to ESI instead of SI)                *
*       - Changed coding style a bit to make code easier to read and added more comments           *
*                                                                                                  *
* 2.11  04.05.2000 SNESAmp v1.2                                                                    *
*       - Control register is now set to 80h on reset                                              *
*                                                                                                  *
* 2.1   04.04.2000 SNESAmp v1.0                                                                    *
*       + Rearranged code so static branch prediction would be more accurate                       *
*       + Clock cycles to emulate is passed on the stack to EmuSPC                                 *
*       + SPC700 registers are passed on the stack to the external debug routine                   *
*       + Eliminated ORG directive and rewrote code to conform to MASM syntax                      *
*       - Fixed some implementation bugs with the speed hack                                       *
*       - Fixed a bug in the 16-bit WrPost macro, again                                            *
*                                                                                                  *
* 2.0   17.03.2000 SNESAmp v0.9                                                                    *
*       + Rewrote for compatibility in a 32-bit C environment                                      *
*       + Added external functions to write to ports                                               *
*       - SLEEP wasn't erasing the port status on the first time through                           *
*                                                                                                  *
* 1.51  22.02.2000 SPC Tool v0.6                                                                   *
*       - Fixed a bug in the 16-bit WrPost macro                                                   *
*                                                                                                  *
* 1.5   01.02.2000 SPC Tool v0.51                                                                  *
*       + Added a hack to eliminate polling the counters                                           *
*       + Aligned many of the short jump destinations on paragraph boundaries                      *
*       + Rewrote the opcode header macro to compile into table form instead of packed form.       *
*         (The compiled code is four times bigger, but it's now easier to align, port, and debug.) *
*       + Store flags as dwords so accesses to the DP don't cause a partial register stall         *
*       - Fixed PCALL, STOP, and SLEEP                                                             *
*                                                                                                  *
* 1.1   21.01.2000 SPC Tool v0.5                                                                   *
*       + Reprogrammed PSW handling for faster entry/exit                                          *
*                                                                                                  *
* 1.0   27.12.1999 SPC Tool v0.2                                                                   *
*                                                      Copyright (C)1999-2004 Alpha-II Productions *
***************************************************************************************************/

#ifndef	__INC_SPC700
#define	__INC_SPC700


//**************************************************************************************************
// Defines

#define	APU_CLK	24576000						//Number of clock cycles in one second

//SPC700 debugging options ---------------------

//Disable DSP data register:
//
//Writes to the DSP data register (0F3h) will have no effect

#define	SPC_NODSP	0x8

//Trace execution:
//
//To step through each instruction.  This flag will cause the debug vector to be called before
// executing the current instruction.
//
//This flag only works if SPC700.Asm is compiled with the SA_DEBUG option enabled (see APU.h).

#define	SPC_TRACE	0x10

//Return immediately from EmuSPC:
//
//This flag only works when used with SPC_TRACE.  Once the debug vector returns, the CPU registers
// are updated then EmuSPC is terminated before the next instruction is executed.
//
//This only returns from EmuSPC, not from EmuAPU.  If you need to terminate DSP emulation as well,
// specify SPC_HALT in addition to SPC_RETURN.

#define	SPC_RETURN	0x1

//Halt APU:
//
//Subsequent calls to EmuSPC will return that cycles were emulated, but nothing will have happened.
//
//Once this flag has been set, the next call to EmuSPC will also disable the DSP emulator.
// Subsequent calls to EmuDSP will generate silence.  This way emulation can be halted in the
// middle of EmuAPU without affecting the size of the output buffer.

#define	SPC_HALT	0x2


//**************************************************************************************************
// Public Variables

//Saved state ----------------------------------
typedef struct SPCState
{
	void	*pRAM;								//[0.0] -> APU's RAM (64k)
	u32		_r1;								//[0.4] reserved
	u16		pc;									//[0.8] Registers
	u8		a,y,x;
	u8		psw;
	u16		sp;

	u32		t8kHz,t64kHz;						//[1.0] # of cycles left until timer increase
	u32		t64Cnt;								//[1.8] # of 64kHz ticks since emulation began
	u32		_r2;								//[1.C] reserved
	u8		upCnt[3];							//[2.0] Up counters for counter increase
	u8		portMod;							//[2.3] Flags for in ports that have been modified
	u8		outp[4];							//[2.4] Out ports
	u32		_r3[2];								//[2.8] reserved
} SPCState;

typedef struct SPCFlags
{
	u8		c:1;								//Carry
	u8		z:1;								//Zero
	u8		i:1;								//Interrupts Enabled (not used in the SNES)
	u8		h:1;								//Half-Carry (auxiliary)
	u8		b:1;								//Software Break
	u8		p:1;								//Direct Page Selector
	u8		v:1;								//Overflow
	u8		n:1;								//Negative (sign)
} SPCFlags;


//**************************************************************************************************
// SPC700 Debugging Routine
//
// A prototype for a function that gets called for debugging purposes
//
// The paramaters passed in can be modified, and on return will be used to update the internal
// registers (except 'cnt').
//
// Note:
//    When modifying PC or SP, only the lower word needs to be modified; the upper word will be
//     ignored.
//
// In:
//    pc -> Current opcode (LOWORD = PC)
//    ya  = YA
//    x   = X
//    psw = PSW
//    sp -> Current stack byte (LOWORD = SP)
//    cnt = Clock cycle counters (four counters, one in each byte)
//          [0-1] 8kHz cycles left until counters 0 and 1 increase
//          [2]   64kHz cycles left until counter 2 increases
//          [3]   CPU cycles left until next 64kHz clock pulse

typedef void SPCDebug(volatile u8 *pc, volatile u16 ya, volatile u8 x,
					  volatile SPCFlags psw, volatile u8 *sp, volatile u32 cnt);

////////////////////////////////////////////////////////////////////////////////////////////////////
// Private Declarations

#ifndef	SNESAPU_DLL

#ifdef	__cplusplus
extern "C"	u8	extraRAM[64];					//RAM used for storage if ROM reading is enabled
extern "C"	u8	outPort[4];						//Four out ports
extern "C"	u32	t64Cnt;							//Counter increased every 64kHz
#else
extern u8	extraRAM[64];
extern u8	outPort[4];
extern u32	t64Cnt;
#endif


////////////////////////////////////////////////////////////////////////////////////////////////////
// External Functions

#ifdef	__cplusplus
extern "C" {
#endif

//**************************************************************************************************
// Initialize SPC700
//
// This function is a remnant from the 16-bit assembly when dynamic code reallocation was used.
// Now it just initializes internal pointers.
//
// Note:
//    You should call InitAPU instead
//
// In:
//    pAPURAM -> 64KB of SPC RAM aligned on a 64K boundary
//
// Destroys:
//    EAX

void InitSPC(void *pAPURAM);


//**************************************************************************************************
// Reset SPC700
//
// Clears all memory, resets the function registers, T64Cnt, and halt flag, and copies ROM into the
// IPL region.
//
// Note:
//    You should call ResetAPU instead
//
// Destroys:
//    EAX

void ResetSPC();


//**************************************************************************************************
// Set SPC700 Debugging Routine
//
// Installs a vector that gets called before instruction execution for debugging purposes
//
// Note:
//    pTrace is always called when a STOP instruction is encountered, regardless of the options.
//    The build option SA_DEBUG must be enabled for pTrace to be called under other circumstances
//
// In:
//    pTrace-> Debugging vector (NULL can be passed to disable the debug vector, -1 leaves the
//             vector currently in place)
//    opts   = SPC700 debugging flags (see SPC_??? defines, -1 leaves the current flags)
//
// Out:
//    Previously installed vector

SPCDebug* SetSPCDbg(SPCDebug *pTrace, u32 opts);


//**************************************************************************************************
// Fix SPC700 After Loading SPC File
//
// Loads timer steps with the values in the timer registers, resets the counters, sets up the in/out
// ports, and stores the registers.
//
// Note:
//    You should call FixAPU instead
//
// In:
//    SPC internal registers
//
// Destroys:
//    EAX

void FixSPC(u16 pc, u8 a, u8 y, u8 x, u8 psw, u8 sp);


//**************************************************************************************************
// Get SPC700 Registers
//
// Returns the registers stored in the CPU
//
// In:
//    -> Vars to store SPC internal registers
//
// Destroys:
//    EAX

void GetSPCRegs(u16 *pPC, u8 *pA, u8 *pY, u8 *pX, u8 *pPSW, u8 *pSP);


//**************************************************************************************************
// Save SPC700 State
//
// Note:
//    Pointers in the SPCState structure can be NULL
//
// In:
//    pState -> Saved state structure
//
// Destroys:
//    EAX

void SaveSPC(SPCState *pState);


//**************************************************************************************************
// Restore SPC700 State
//
// Note:
//    Setting a pointer in SPCState to NULL will keep its corresponding internal data intact.
//
// In:
//    pState -> Saved state structure
//
// Destroys:
//    EAX

void RestoreSPC(SPCState *pState);


//**************************************************************************************************
// Write to APU RAM
//
// Writes a value to APU RAM.  Use this instead of writing to RAM directly so any necessary internal
// changes can be made.
//
// In:
//    addr = Address to write to (only the lower 16-bits are used)
//    val  = Value to write
//
// Destroys:
//    EAX

void SetAPURAM(u32 addr, u8 val);


//**************************************************************************************************
// Write to SPC700 Port(s)
//
// Writes a value to the SPC700 via the in ports.  Use this instead of writing to RAM directly.
// SPCIn writes a single byte value while SPCInW writes a 16-bit value to two ports.
//
// In:
//    port = Port on which to write (0-3)
//    val  = Value to write
//
// Destroys:
//    EAX

void InPort(u32 port, u8 val);
void InPortW(u32 port, u16 val);


//**************************************************************************************************
// Emulate SPC700
//
// Emulates the SPC700 for the number of clock cycles specified, or if the counter break option is
// enabled, until a counter is increased, whichever happens first.
//
// Note:
//    You should call EmuAPU instead
//    Passing values <= 0 will cause undeterminable results
//
// In:
//    cyc = Number of 24.576MHz clock cycles to execute (must be > 0)
//
// Out:
//    Clock cycles left to execute (negative if more cycles than specified were emulated)
//
// Destroys:
//    nothing

s32 EmuSPC(s32 cyc);

#ifdef	__cplusplus
}
#endif

#endif	//SNESAPU_DLL
#endif	//__INC_SPC700
