;
;Program:    Super Nintendo Entertainment System(tm) Audio Processing Unit Emulator
;Platform:   Intel 80386 & MMX
;Programmer: Anti Resonance
;
;"SNES" and "Super Nintendo Entertainment System" are trademarks of Nintendo Co., Limited and its
;subsidiary companies.
;
;This program is free software; you can redistribute it and/or modify it under the terms of the
;GNU General Public License as published by the Free Software Foundation; either version 2 of
;the License, or (at your option) any later version.
;
;This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
;without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
;See the GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License along with this program;
;if not, write to the Free Software Foundation, Inc.
;59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;
; Revision History:
;
; 1.5  04.11.2003 SNESAPU 2.0
;      + Updated EmuAPU to use SetEmuDSP when DSPINTEG is enabled
;      + Move fade out code into DSP.Asm
;
; 1.41 12.07.2003 SNESAPU 1.0a
;      - Fixed the call to SetDSPVol in SetFade that wasn't restoring the stack
;
; 1.4  20.06.2003
;      + Added LoadSPCFile
;      + Added ability to pass -1 to ResetAPU to keep current amp level
;
; 1.3  10.02.2003
;      + Rewrote in assembly
;      + Removed GetAPUData and memory allocation (they were specific to SNESAPU.DLL)
;      - Sometimes EmuAPU would generate an extra sample
;
; 1.22 20.06.2002
;      - Call SetFade after seeking
;
; 1.21 12.04.2001
;      + Updated SeekAPU to incorporate new features in SPC700 and DSP emus
;
; 1.2  30.12.2000 SNESAmp v2.0
;      + Added SetAPULength, SeekAPU, SaveAPU, RestoreAPU, HaltAPU
;      + General optimizations
;      - Simplified interface a bit
;
; 1.1  04.04.2000 SNESAmp v1.0
;
; 1.0  17.03.2000 SNESAmp v0.9
;                                                          Copyright (C)2003-04 Alpha-II Productions
;

%define APU_INC

;
;Options

CPU_CYC	EQU	24										;CPU clock divisor
															;10 = 2.457MHz
															;12 = 2.048MHz
															;24 = 1.024MHz

DEBUG		EQU	1										;Enable debugging ability
DSPINTEG	EQU	1										;Integrate DSP emulation with SPC700

HALFC		EQU	0										;Enable half-carry flag emulation for decimal adjusts
CNTBK		EQU	0										;Break SPC700/Update DSP if a counter increases
SPEED		EQU	1										;Enable speed hacks for counter reads and SLEEP
IPLW		EQU	1										;Enables the IPL Writable flag in register F1
DSPBK		EQU	1										;Break SPC700/Update DSP if register F3 is read

VMETERM	EQU	1										;Monitor main output level (for APR)
VMETERV	EQU	1										;Monitor levels of each voice (for visualization)
SNESINT	EQU	1										;Use a gaussian curve generated by the SNES, otherwise
															; create a more accurate curve at run-time
STEREO	EQU	1										;Enable stereo controls (separation and EFBCT)


;
;Public Variables

PUBLIC	apuOpt										;Build options


;
;Exported Functions

;
;Initialize Audio Processing Unit
;
;Initializes internal tables, mixer settings, memory, and registers.
;This only needs to be called once, and must be proceeded by ResetAPU.
;
;In:
;   pAPURAM -> 64kB of memory aligned on a 64kB boundary

PUBLIC InitAPU, pAPURAM:ptr


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Reset Audio Processor
;
;Clears all memory, sets registers to default values, and sets the amplification level.
;
;In:
;   amp = Amplification (-1 = keep current amp level, see SetDSPAmp for more information)

PUBLIC ResetAPU, amp:dword


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Fix Audio Processor After Load
;
;Prepares the sound processor for emulation after an .SPC/.ZST is loaded.
;
;In:
;   SPC700 internal registers

PUBLIC FixAPU, pc:word, a:byte, y:byte, x:byte, psw:byte, s:byte


;
;Load SPC File
;
;Restores the APU state from an SPC file.  This eliminates the need to call ResetAPU, copy memory,
;and call FixAPU.
;
;In:
;   pFile -> 66048 byte SPC file

PUBLIC LoadSPCFile, pFile:dword


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Save Audio Processor State
;
;Creates a saved state of the audio processor
;
;Note:  Pointers within the structures can be NULL
;
;In:
;   pSPC -> SPCState structure
;   pDSP -> DSPState structure

PUBLIC SaveAPU, pSPC:ptr, pDSP:ptr


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Restore Audio Processor State
;
;Restores the audio processor from a saved state
;
;Notes:  Pointers within the structure can be NULL
;        Passing a NULL value in pEchoBuf will cause the echo to be reset
;        You do not need to call FixAPU after RestoreAPU
;
;In:
;   pSPC -> SPCState structure
;   pDSP -> DSPState structure

PUBLIC RestoreAPU, pSPC:ptr, pDSP:ptr


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Set Audio Processor Options
;
;Configures the sound processor emulator.  Range checking is performed on all parameters.
;
;Notes:  -1 can be passed for any parameter you want to remain unchanged
;        see SetDSPOpt() in DSP.Inc for a more detailed description
;
;In:
;   mix   = Mixing routine (default MIX_INT)
;   chn   = Number of channels (1 or 2, default 2)
;   bits  = Sample size (8, 16, 24, 32, or -32 [IEEE 754], default 16)
;   rate  = Sample rate (8000-192000, default 32000)
;   inter = Interpolation type (default INT_GAUSS)
;   opts  = See 'DSP options' in the Defines section of DSP.Inc

PUBLIC SetAPUOpt, mix:dword, chn:dword, bits:dword, rate:dword, inter:dword, opts:dword


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Set Audio Processor Sample Clock
;
;Calculates the ratio of emulated clock cycles to sample output.  Used to speed up or slow down a
;song without affecting the pitch.
;
;In:
;   speed = Multiplier [16.16] (1/2x to 16x)

PUBLIC SetAPUSmpClk, speed:dword


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Set Audio Processor Song Length
;
;Sets the length of the song and fade
;
;Notes:  If a song is not yet playing, you must call ResetAPU or set T64Cnt to 0 before calling this
;        To set a song with no length, pass -1 and 0 for the song and fade
;
;In:
;   song = Length of song (in 1/64000ths second)
;   fade = Length of fade (in 1/64000ths second)
;
;Out:
;   EAX = Total length

PUBLIC SetAPULength, song:dword, fade:dword


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Emulate Audio Processing Unit
;
;Emulates the APU for a specified amount of time.  DSP output is placed in a buffer to be handled
;by the main program.
;
;In:
;   pBuf-> Buffer to store output samples
;   len  = Length of time to emulate (must be > 0)
;   type = Type of parameter passed in len
;          0 - len is the number of APU clock cycles to emulate (APU_CLK = 1 second)
;          1 - len is the number of samples to generate
;Out:
;   EAX -> End of buffer

PUBLIC EmuAPU, pBuf:ptr, len:dword, type:byte


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Seek to Position
;
;Seeks forward in the song
;
;In:
;   time = 1/64000ths second to seek forward
;   fast = Use faster seeking method (may break some songs)

PUBLIC SeekAPU, time:dword, fast:bool


;AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
;Shutdown Audio Processing Unit
;
;Currently does nothing

PUBLIC ShutAPU, NULL